/*

	bmp.c - bitmap support routines for the Power JBIG-2 Decoder / Encoder


	Author:
		Dave Tompkins (dave.tompkins@ieee.org)
		University of British Columbia

	Source:
		http://spmg.ece.ubc.ca/jbig2
	  
	Revision History:

		1.0.0	1999-03-15	First Release


	Copyright:
		(c) 1998 University of British Columbia and Image Power, Inc.

		The program may be used for academic, demonstration and research 
		purposes only.  It may not be sold for profit or incorporated into
		any commercial programs and/or services without first obtaining a
		commercial useage license from Image Power, Inc.  This program is
		provided as is, without any express or implied warranty, including
		the warranty of fitness for a particular purpose or warranties of
		merchantability.

*/

#include "jb2common.h"

/* ConstMask is used to check specific bit positions */

BYTE ConstMask1[8] = {128,64,32,16,8,4,2,1};
BYTE ConstMask0[8] = {127,191,223,239,247,251,253,254};

/* AreaVal is used to determine the number of 1's in a BYTE */

static const ULONG AreaVal[256] = {
	0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

/* AreaBMP calculates the number of 1's in the entire image */

ULONG AreaBMP(BMP *image) {
	ULONG x;
	ULONG y;
	ULONG ret = 0;
	BYTE mask = 0xFF;
	if (MOD8(image->Width))
		mask <<= (8-MOD8(image->Width));
	for (y=0;y<image->Height;y++) {
		for (x=0;x<image->MinWidthBytes-1;x++)
			ret += AreaVal[image->Bits[y*image->AlignWidthBytes + x]];
		ret += AreaVal[mask & image->Bits[y*image->AlignWidthBytes + image->MinWidthBytes - 1]];
	}
	return(ret);
}

/* CreateBMP initializes & allocates memory for a new bitmap */

void CreateBMP(BMP *newBMP,ULONG w,ULONG h,ULONG def) {

	newBMP->Width = w;
	
	/* if there is striped or length-unknown data, use 1 */
	if (h==0xFFFFFFFF)
		h = 1;
	
	newBMP->Height = h;
	
	newBMP->MinWidthBytes = w / 8;
	if (MOD8(newBMP->Width))
		newBMP->MinWidthBytes++;

	newBMP->AlignWidthBytes = newBMP->MinWidthBytes;
	while (newBMP->AlignWidthBytes % BMPALIGNBYTES)
		newBMP->AlignWidthBytes++;

	newBMP->Bits = (BYTE *) jb2calloc(newBMP->AlignWidthBytes * h,sizeof(BYTE));

	if (def==1)
		InvertBMP(newBMP);
}

/* For every CreateBMP, DestroyBMP should be called to free() the memory */

void DestroyBMP (BMP *oldBMP) {
	jb2free(oldBMP->Bits);
	oldBMP->Bits = NULL;
}

/* CombineByte returns the value of (a COMBOP b)
   but for only the bits set to 0 in mask */

static BYTE CombineByte(BYTE a, BYTE b, ULONG op,BYTE mask) {
	
	static BYTE result;

	switch (op) {
		case COMBOPOR:
			result = (BYTE)(a|b);
			break;
		case COMBOPAND:
			result = (BYTE)(a&b);
			break;
		case COMBOPXOR:
			result = (BYTE)(a^b);
			break;
		case COMBOPXNOR:
			result = (BYTE)(~(a^b));
			break;
		case COMBOPREPLACE:
			result = b;
	}

	if (mask==0)
		return (result);
	else
		return ((BYTE)((a&(~mask))|(result&mask)));
}

/* CombineBMP is a fairly messy routine which combines the bitmap sub
   onto the bitmap main with operartor op , at co-ordinate (ox,oy)
   Most of the code deals with the problem of un-aligned bitmaps */

void CombineBMP(BMP *main,BMP *sub,SLONG ox,SLONG oy,ULONG op) {

	SLONG x;
	SLONG y;
	SLONG loffset = 0;
	SLONG roffset;
	ULONG numbyte;
	ULONG loff8;
	ULONG roff8;
	SLONG oxdiv8;
	SLONG oxldiv8;
	SLONG rdiff;
	SLONG Xmin;
	SLONG Xmax;
	SLONG Ymin;
	SLONG Ymax;
	BYTE *SubBits;
	BYTE *MainBits;
	BYTE mask;
	BYTE newbyte;

	Ymin = MAX(0,-oy);
	Ymax = MIN((SLONG)sub->Height,(SLONG)main->Height-oy);

	if (MOD8(ox)!=0) {
		loffset=(8-MOD8(ox));
		if (loffset>8)
			loffset -= 8;
	}
	loff8 = 8 - loffset;
	oxdiv8 = ox / 8;
	oxldiv8 = (ox+loffset) / 8;

	roffset = (SLONG)(MOD8(sub->Width+ox));
	if (roffset<0)
		roffset = 0;
	roff8 = 8 - roffset;

	numbyte = (sub->Width - loffset - roffset) / 8;

	/* Special Case: Sub is very narrow */
	if ((ULONG)(loffset + roffset)>sub->Width) { 
		if ((ox>=0)&&(oxdiv8<(SLONG)main->MinWidthBytes)) {
			mask = (BYTE)((0xFF>>loff8)&(0xFF<<roff8));
			SubBits = sub->Bits + Ymin * sub->AlignWidthBytes;
			MainBits = main->Bits + (Ymin+oy) * main->AlignWidthBytes + oxdiv8;
			for (y=Ymin;y<Ymax;y++) {
				*MainBits = CombineByte(*MainBits,(BYTE)(*SubBits>>loff8),op,mask);
				SubBits += sub->AlignWidthBytes;
				MainBits += main->AlignWidthBytes;
			}
		}
		loffset = 0;
		numbyte = 0;
		roffset = 0;
	}

	if (loffset)
		if ((ox>=0)&&(oxdiv8<(SLONG)main->MinWidthBytes)) {
			mask = (BYTE)((0xFF >> loff8));
			SubBits = sub->Bits + Ymin * sub->AlignWidthBytes;
			MainBits = main->Bits +(Ymin+oy) * main->AlignWidthBytes + oxdiv8;
			for (y=Ymin;y<Ymax;y++) {
				*MainBits = CombineByte(*MainBits,(BYTE)((*SubBits)>>loff8),op,mask);
				SubBits += sub->AlignWidthBytes;
				MainBits += main->AlignWidthBytes;
			}
		}

	if (numbyte) {
		Xmin = MAX(0,-oxldiv8);
		Xmax = MIN((SLONG)numbyte,(SLONG)main->MinWidthBytes-oxldiv8);
		for (y=Ymin;y<Ymax;y++) {
			SubBits = sub->Bits + y * sub->AlignWidthBytes + Xmin;
			MainBits = main->Bits +(y+oy) * main->AlignWidthBytes + Xmin + oxldiv8;
			for (x=Xmin;x<Xmax;x++) {
				if (loffset)
					newbyte = (BYTE)((((*SubBits)<<8)+*(SubBits+1))>>loff8);
				else
					newbyte = (*SubBits);
				*MainBits = CombineByte(*MainBits,newbyte,op,0);
				MainBits++;
				SubBits++;
			}
		}
	}

	if (roffset)
		if (((SLONG)numbyte+oxldiv8>=0)&&((SLONG)numbyte+oxldiv8<(SLONG)main->MinWidthBytes)) {
			mask = (BYTE)(0xFF << roff8);

			rdiff = (SLONG)(MOD8(sub->Width) - roffset);
			if (MOD8(sub->Width) == 0)
				rdiff += 8;

			SubBits = sub->Bits + (Ymin)*sub->AlignWidthBytes + sub->MinWidthBytes - 1;
			MainBits = main->Bits +(Ymin+oy) * main->AlignWidthBytes + numbyte + oxldiv8;
			for (y=Ymin;y<Ymax;y++) {
				if (rdiff < 0)
					newbyte = (BYTE)(((*(SubBits-1)<<8)+*SubBits)>>(-rdiff));
				else
					newbyte = (BYTE)(*SubBits << rdiff);
				*MainBits = CombineByte(*MainBits,newbyte,op,mask);
				SubBits += sub->AlignWidthBytes;
				MainBits += main->AlignWidthBytes;
			}
		}
}

/* WriteBytesToFiel ensures the proper byte ordering for both Win & Unix */

void WriteBytesToFile(FILE *filOUT,ULONG val, ULONG numbytes) {

	ULONG j;

	for (j=0;j<numbytes;j++) {
		fputc((BYTE)(val & 0xFF),filOUT);
		val >>= 8;
	}

}

/* SaveBMP saves the bitmap to a file, using the standard Windows .BMP format */

void SaveBMP(const char *pFileName, BMP *image) {
	FILE *filOut;
	ULONG x;
	ULONG y;
	ULONG j;
	ULONG offset = 0;

	if (image->MinWidthBytes % 4 > 0)
		offset = 4 - image->MinWidthBytes % 4;

	filOut = fopen(pFileName, "wb");

	WriteBytesToFile(filOut,0x42,1);
	WriteBytesToFile(filOut,0x4d,1);
	WriteBytesToFile(filOut,image->Height * (image->MinWidthBytes + offset) + 62,4);
	WriteBytesToFile(filOut,0,4);
	WriteBytesToFile(filOut,62,4);
	WriteBytesToFile(filOut,0x28,4);
	WriteBytesToFile(filOut,image->Width,4);
	WriteBytesToFile(filOut,image->Height,4);
	WriteBytesToFile(filOut,1,2);
	WriteBytesToFile(filOut,1,2);
	WriteBytesToFile(filOut,0,4);
	WriteBytesToFile(filOut,0,4);
	WriteBytesToFile(filOut,0,4);
	WriteBytesToFile(filOut,0,4);
	WriteBytesToFile(filOut,2,4);
	WriteBytesToFile(filOut,2,4);
	WriteBytesToFile(filOut,0x00FFFFFF,4);
	WriteBytesToFile(filOut,0,4);

	for (y=0;y<image->Height;y++) {
		for(x=0;x<image->MinWidthBytes;x++)
			fputc(image->Bits[(image->Height-y-1)*image->AlignWidthBytes+x],filOut);
		for (j=0;j<offset;j++)
			fputc(0x00,filOut);
		}

	fclose(filOut);

}

/* LoadBMP loads a bitmap from a file, using the standard Windows .BMP format
   Notes:
   * only monochrome images should be loaded by LoadBMP
   * The image is checked to see if 0= black or white
*/

ULONG LoadBMP(const char *pFileName, BMP *image) {

	FILE *filIn;
	ULONG x;
	ULONG y;
	ULONG j;
	ULONG w;
	ULONG h;
	ULONG offset = 0;
	BYTE filHeader[62];

	filIn = fopen(pFileName, "rb");

	if (filIn==NULL)
		return(1);
	else {
	
		for (j=0;j<62;j++) {
				filHeader[j] = (BYTE)(fgetc(filIn));
				if (feof(filIn))
					LogError("Unexpected EOF in input file");
		}

		if ((filHeader[0]!=0x42)||(filHeader[1]!=0x4d))
			LogError("Invalid .BMP file format!");

		if ((filHeader[28]!=1)||(filHeader[29]!=0))
			LogError("Invalid .BMP, Incorrect Bits Per Pixel");

		w = filHeader[21];
		w = (w << 8) + filHeader[20];
		w = (w << 8) + filHeader[19];
		w = (w << 8) + filHeader[18];
		h = filHeader[25];
		h = (h << 8) + filHeader[24];
		h = (h << 8) + filHeader[23];
		h = (h << 8) + filHeader[22];

		CreateBMP(image,w,h,0);

		if (image->MinWidthBytes % 4 > 0)
			offset = 4 - image->MinWidthBytes % 4;

		for (y=0;y<image->Height;y++) {
			for(x=0;x<image->MinWidthBytes;x++)
				image->Bits[(image->Height-y-1) * image->AlignWidthBytes + x] = (BYTE)(fgetc(filIn));
			for (j=0;j<offset;j++)
				fgetc(filIn);
		}

		fclose(filIn);

		if ((filHeader[54]==0)&&(filHeader[58]==0xFF))
			InvertBMP(image);
		
		return(0);
	}
}

/* A handy routine for debugging & viewing results */

void printfBMP(BMP *img) {

	ULONG x;
	ULONG y;

	for (y = 0;y < img->Height; y++) {
		for (x = 0; x < img->Width; x++) {
			(GetBitBMP(img,x,y)) ? printf("O") : printf(".");
		}
		printf("\n");
	}
	printf ("\n");
}

/* Extract Region extracts the rectangle New from Old, leaving Old unchanged */

void ExtractRegion (BMP *New, BMP *Old, ULONG ox, ULONG oy) {
	CombineBMP(New,Old,- (SLONG) ox,- (SLONG) oy,COMBOPREPLACE);
}

/* ClearRegion is used to set a rectangular region of a bitmap to 0 */

void ClearRegion (BMP *Image, ULONG ox, ULONG oy, ULONG width, ULONG height) {
	BMP dummy;
	CreateBMP(&dummy,width,height,0);
	CombineBMP(Image,&dummy,ox,oy,COMBOPREPLACE);
	DestroyBMP(&dummy);
}

/* CopyBMP creates a new instance of an existing BMP */

void CopyBMP (BMP *New, BMP *Old) {
	CreateBMP(New,Old->Width,Old->Height,0);
	CombineBMP(New,Old,0,0,COMBOPOR);
}

/* ExactBMP is used to determine of two bitmaps are identical */

ULONG ExactBMP(BMP *a, BMP *b) {
	ULONG k;
	ULONG l;
	BYTE mask;

	if ((a->Width==b->Width)&&(a->Height==b->Height)) { 
		if (MOD8(a->Width)==0)
			mask = 0xFF;
		else
			mask = (BYTE)(0xFF << (8 - MOD8(a->Width)));

		for (k = 0; k < a->Height; k++) {
			for (l = 0; l < a->MinWidthBytes - 1; l++)
				if (a->Bits[k * a->AlignWidthBytes + l]!=b->Bits[k * a->AlignWidthBytes + l])
					return(0);
			if ((mask&(a->Bits[k * a->AlignWidthBytes + l]))!=(mask&(b->Bits[k * a->AlignWidthBytes + l])))
				return(0);
		}
		return(1);
	} else
		return(0);
}

/* InvertBMP inverts the colours (white=black, black=white) of an image */

void InvertBMP (BMP *Image) {
	ULONG x;
	ULONG y;
	for (y=0;y<Image->Height;y++)
		for (x=0;x<Image->MinWidthBytes;x++)
			Image->Bits[y*Image->AlignWidthBytes + x] = (BYTE)(~Image->Bits[y*Image->AlignWidthBytes + x]);
}

