/*

	deccommon.c - decoder support routines for the Power JBIG-2 Decoder

	Author:
		Dave Tompkins (dave.tompkins@ieee.org)
		University of British Columbia

	Source:
		http://spmg.ece.ubc.ca/jbig2
	  
	Revision History:

		1.0.0	1999-03-15	First Release


	Copyright:
		(c) 1998 University of British Columbia and Image Power, Inc.

		The program may be used for academic, demonstration and research 
		purposes only.  It may not be sold for profit or incorporated into
		any commercial programs and/or services without first obtaining a
		commercial useage license from Image Power, Inc.  This program is
		provided as is, without any express or implied warranty, including
		the warranty of fitness for a particular purpose or warranties of
		merchantability.

*/

#include "jb2common.h"
#include "deccommon.h"

/* The callback function for the MQ Decoder */

BYTE MQMainIn(void *pDecoder) {
	return ((BYTE) DecodeByte((DECODER *) pDecoder));
}

/* Returns the pointer to the current decoder position */

BYTE *DecoderPtr(DECODER *pD) {
	return (pD->pDec);
}

/* Determines if the decoder has extended beyond the number of bits */

BOOLN DecodeEOS(DECODER *pD) {
	if (pD->pDec>=pD->pDecEOF)
		return(1);
	else
		return(0);
}

/* Safely fetches the next byte for the decoder */

void DecNextByte(DECODER *pD) {
	if (pD->pDec==pD->pDecEOF) {
		LogError("DecNextByte() - Input past the End of Stream");
		pD->DecCurrent = 0xFF;
	} else {
		pD->DecCurrent = *pD->pDec;
		pD->pDec++;
	}
	pD->DecBitCount = 8;
	pD->ByteCount ++;
}

/* DecodeBits returns the next N bytes from the decoder */

ULONG DecodeBits(DECODER *pD,ULONG N) {
	ULONG ret = 0;

	if (N==0)
		return(0);
	
	if (pD->DecBitCount == 0)
		DecNextByte(pD);
	
	if (N<=pD->DecBitCount) {
		ret = pD->DecCurrent >> (8 - N);
		pD->DecCurrent <<= N;
		pD->DecBitCount -= N;
	} else {
		ret = pD->DecCurrent >> (8 - pD->DecBitCount);
		N -= pD->DecBitCount;
		while (N >= 8) {
			DecNextByte(pD);
			ret = (ret << 8) + pD->DecCurrent;
			N -= 8;
			pD->DecBitCount = 0;
		}
		if (N>0) {
			DecNextByte(pD);
			ret = (ret << N) + (pD->DecCurrent >> (8 - N));
			pD->DecCurrent <<= N;
			pD->DecBitCount -= N;
		}
	}
	return(ret);
}

/* DecodeFlush forces the decoder to the next byte boundary */

void DecodeFlush(DECODER *pD) {
	pD->DecBitCount = 0;				 
}

/* InitDecoder sets up a decoder and its' boundaries */

void InitDecoder(DECODER *pD,BYTE *start, ULONG length) {
	pD->DecBitCount = 0;
	pD->ByteCount = 0;
	pD->pDec = start;
	pD->pDecEOF = start + length;
}

/* Forces the decoder to skip over (or rewind) a specified # of bits */

void DecSkipBytes(DECODER *pD,SLONG numskip) {
	DecodeFlush(pD);
	pD->pDec += numskip;
}

/* returns the Identifier for the leaf of the Huffman Tree when decoded */

ULONG DecodeHuffTree (DECODER *pDecoder, HUFFNODE *mainnode) {

	HUFFNODE *pCurNode;

	pCurNode = mainnode;

	while (pCurNode->bEND==0) {
		if(DecodeBit(pDecoder))
			pCurNode=pCurNode->p1;
		else
			pCurNode=pCurNode->p0;
		
		if (pCurNode==NULL) {
			LogError("Decoding Huffman Code: No Prefix Match!");
			return(0);
		}
	}
	return(pCurNode->NodeID);
}

/* returns the value and OOB flag from the next huffman code in the bitstream */

SLONG DecodeHuff(DECODER *pD,HUFFTABLE *tbl,ULONG *bOOB) {

	ULONG RangeCode;
	ULONG idx;
	SLONG ret = 0;
	
	*bOOB = 0;

	idx = DecodeHuffTree(pD,tbl->Tree);
	
	RangeCode = DecodeBits(pD,tbl->RANGELEN[idx]);

	if (idx==tbl->pDef->numLines + 2)
		*bOOB = 1;
	else if (idx==tbl->pDef->numLines)
		ret = tbl->RANGELOW[idx] - RangeCode;
	else
		ret = tbl->RANGELOW[idx] + RangeCode;

	return (ret);
}

/* These dummy functions are included to ensure proper compliation of shared files */

void EncodeBits(ENCODER *pE, ULONG a, ULONG b) {
	pE = pE;
	a = b = LogWarning ("EncodeBits() called from Decoder");
}
void EncodeFlush(ENCODER *pE) {
	pE = pE;
	LogWarning ("EncodeFlush called from Decoder");
}