/*

	decsegments.c - General Segment routines for the Power JBIG-2 Decoder


	Author:
		Dave Tompkins (dave.tompkins@ieee.org)
		University of British Columbia

	Source:
		http://spmg.ece.ubc.ca/jbig2
	  
	Revision History:

		1.0.0	1999-03-15	First Release


	Copyright:
		(c) 1998 University of British Columbia and Image Power, Inc.

		The program may be used for academic, demonstration and research 
		purposes only.  It may not be sold for profit or incorporated into
		any commercial programs and/or services without first obtaining a
		commercial useage license from Image Power, Inc.  This program is
		provided as is, without any express or implied warranty, including
		the warranty of fitness for a particular purpose or warranties of
		merchantability.

*/

#include "jb2common.h"
#include "deccommon.h"

/* Determine which routine should be called, based upon the segment type */

void DecodeSegment (SEGMENTHEADER *Seg) {

	InitDecoder(&Seg->pCoderData->Dec,Seg->pData,Seg->DataLength);

	switch (Seg->Type) {
		case 0: /* Symbol Dictionary */
			DecodeSymbolDictionarySegment(Seg);
			break;
		case 4: /* Intermediate Symbol Region */
		case 6: /* Immediate Symbol Region */
		case 7: /* Immediate Lossless Symbol Region */
			DecodeSymbolRegionSegment(Seg);
			break;
		case 16: /* HalfTone Dictionary */
			DecodeHalftoneDictionarySegment(Seg);
			break;
		case 20: /* Intermediate HalfTone Region */
		case 22: /* Immediate HalfTone Region */
		case 23: /* Immediate LossLess HalfTone Region */
			DecodeHalftoneRegionSegment(Seg);
			break;
		case 36: /* Intermediate Generic Regions */
		case 38: /* Immediate Generic Regions */
		case 39: /* Immediate Lossless Generic Regions */
			DecodeGenericRegionSegment(Seg);
			break;
		case 40: /* Intermediate refinement region */
		case 42: /* Immediate refinement region */
		case 43: /* Immediate Lossless refinement region */
			DecodeRefinementRegionSegment(Seg);
			break;
		case 48: /* Page Information */
			break;
		case 49: /* End of Page */
			break;
		case 50: /* End of Stripe */
			break;
		case 51: /* End of File */
			break;
		case 52: /* Supported Profiles */
			break;
		case 53: /* Tables */
			DecodeTableSegment(Seg);
			break;
		case 62: /* Extension */
			break;	
	}
}

/* Create the structures & data for the JB2File structure */

BOOLN JB2ParseFile (JB2FILE *pJB2File) {

	ULONG j,k;

	ULONG maxPages;
	ULONG maxAssoc;
	ULONG maxSegments = 16; /* Note: This is Dynamic, 16 is Default */	

	BYTE FileHeader[11];
	BOOLN bRandomAccess;
	ULONG expectedPages;
	
	DECODER Dec;
	
	pJB2File->numSegments = 0;
	
	pJB2File->SegmentHeaders = jb2malloc(maxSegments * sizeof(*pJB2File->SegmentHeaders));

	InitDecoder(&Dec,pJB2File->pFile,pJB2File->FileLength);

	/* determine that it _is_ a .JB2 file we recognize */
	for (j=0;j<8;j++)
		FileHeader[j] = (BYTE) DecodeByte(&Dec);

	if (ValidFileHeader(FileHeader)==0) {
		LogError("Invalid File Header");
		return(1);
	}

	/* determine if Random or sequential access file */
	FileHeader[9] = (BYTE) DecodeByte(&Dec);
	bRandomAccess = !(getBIT(FileHeader[9],0));

	/* determine the expected # of pages (not used) */	
	if (getBIT(FileHeader[9],1)==0)
		expectedPages = DecodeQuad(&Dec);
	else
		expectedPages = 0;

	if (bRandomAccess) {
		/* read in all segment headers until the EOF segment is encountered */
		do 
			ReadSegmentHeader(&Dec, &pJB2File->SegmentHeaders, &pJB2File->numSegments, &maxSegments);
		while (pJB2File->SegmentHeaders[pJB2File->numSegments - 1].Type != 51);

		/* determine where each data segment */
		for (j=0;j<pJB2File->numSegments;j++) {
			pJB2File->SegmentHeaders[j].pData = DecoderPtr(&Dec);
			if (pJB2File->SegmentHeaders[j].bUnknownLength) {
				SkipUnknownSegmentLength(&Dec);
				pJB2File->SegmentHeaders[j].RegSegment.Height = DecodeQuad(&Dec);
				pJB2File->SegmentHeaders[j].DataLength = DecoderPtr(&Dec) - pJB2File->SegmentHeaders[j].pData;
			}	
			else
				DecSkipBytes(&Dec,pJB2File->SegmentHeaders[j].DataLength);
		}
	} else
		while (!DecodeEOS(&Dec)) {
			ReadSegmentHeader(&Dec, &pJB2File->SegmentHeaders, &pJB2File->numSegments, &maxSegments);
			pJB2File->SegmentHeaders[pJB2File->numSegments - 1].pData = DecoderPtr(&Dec);
			if (pJB2File->SegmentHeaders[pJB2File->numSegments - 1].bUnknownLength) {
				SkipUnknownSegmentLength(&Dec);
				pJB2File->SegmentHeaders[j].RegSegment.Height = DecodeQuad(&Dec);
				pJB2File->SegmentHeaders[j].DataLength = DecoderPtr(&Dec) - pJB2File->SegmentHeaders[j].pData;
			}	
			else
				DecSkipBytes(&Dec, pJB2File->SegmentHeaders[pJB2File->numSegments - 1].DataLength);
		}

	/* Pass 0: Assign Global Coder data*/

	MQSetInFunction(&pJB2File->CoderData.MQDec,&MQMainIn,(void *) &pJB2File->CoderData.Dec);
	CreateMQBuffer(&pJB2File->CoderData.Generic,MQCONTEXTMAX);
	CreateMQBuffer(&pJB2File->CoderData.Refine,MQCONTEXTMAX);
	InitializeMMR(&pJB2File->CoderData.MMRTables);
	InitializeHuffStandard(pJB2File->CoderData.aHuffStd);

	for (j=0;j<pJB2File->numSegments;j++)	
		pJB2File->SegmentHeaders[j].pCoderData = &pJB2File->CoderData;

	/* Pass 1: Determine the number of pages */
	maxPages = 0;
	for (j=0;j<pJB2File->numSegments;j++)
		if (pJB2File->SegmentHeaders[j].Type == 48)
			maxPages ++;

	pJB2File->Pages = jb2malloc(maxPages * sizeof(*pJB2File->Pages));

	/* Pass 2: Read Page Segment Headers */
	pJB2File->numPages = 0;
	for (j=0;j<pJB2File->numSegments;j++)
		if (pJB2File->SegmentHeaders[j].Type == 48) {
			pJB2File->SegmentHeaders[j].pPageSeg = &pJB2File->Pages[pJB2File->numPages].PageSeg;
			DecodePageSegment(&pJB2File->SegmentHeaders[j]);
			pJB2File->Pages[pJB2File->numPages].Number = pJB2File->SegmentHeaders[j].PageNum;
			if (pJB2File->SegmentHeaders[j].PageNum==0)
				LogWarning("Cannot have a Page #0");
			pJB2File->numPages ++;
		}

	/* Pass 3: Assign Segments to Pages & Build Array of Segments For Each Page */
	for (j=0;j<pJB2File->numPages;j++) {
		maxAssoc = 0;
		/* Pass 3.1: Determine # of segments for Page j & assign segs to j*/
		for (k=0;k<pJB2File->numSegments;k++)
			if (pJB2File->SegmentHeaders[k].PageNum == pJB2File->Pages[j].Number) {
				maxAssoc++;
				pJB2File->SegmentHeaders[k].pPageImage = &pJB2File->Pages[j].PageImage;
				pJB2File->SegmentHeaders[k].pPageSeg = &pJB2File->Pages[j].PageSeg;
		}
		/* Pass 3.2: Build Array of segments for Page j */
		pJB2File->Pages[j].apSegments = jb2malloc(maxAssoc * sizeof(*pJB2File->Pages[j].apSegments));
		pJB2File->Pages[j].numSeg = 0;
		for (k=0;k<pJB2File->numSegments;k++)
			if (pJB2File->SegmentHeaders[k].PageNum == pJB2File->Pages[j].Number) {
				pJB2File->Pages[j].apSegments[pJB2File->Pages[j].numSeg] = &pJB2File->SegmentHeaders[k];
				pJB2File->Pages[j].numSeg ++;
			}
	}

	BuildSegmentTree(pJB2File->SegmentHeaders,pJB2File->numSegments);

	return(0);
}

/* Decode into memory all the Segments with a page = 0 */
void JB2DecodeGlobal(JB2FILE *pJB2File) {
	ULONG j;
	for (j=0;j<pJB2File->numSegments;j++)
		if (pJB2File->SegmentHeaders[j].PageNum==0)
			DecodeSegment(&pJB2File->SegmentHeaders[j]);
}

/* Decode a Page by decoding each of the segments associated with that page */

void JB2DecodePage(JB2PAGE *pJB2Page) {
	ULONG j;
	for (j=0;j<pJB2Page->numSeg;j++)
		DecodeSegment(pJB2Page->apSegments[j]);
}

/* frees up any existing JB2File structures and any memory leftovers */

void JB2FreeFile(JB2FILE *pJB2File) {
	ULONG j;

	DestroyMQBuffer(&pJB2File->CoderData.Generic);
	DestroyMQBuffer(&pJB2File->CoderData.Refine);
	FreeMMR(&pJB2File->CoderData.MMRTables);
	FreeHuffStandard(pJB2File->CoderData.aHuffStd);

	for (j=0;j<pJB2File->numSegments;j++)
		FreeSegmentHeader(&pJB2File->SegmentHeaders[j]);

	for (j=0;j<pJB2File->numPages;j++)
		jb2free(pJB2File->Pages[j].apSegments);

	jb2free(pJB2File->SegmentHeaders);
	jb2free(pJB2File->Pages);
}

/* The SegmentTree finds all the dependencies and referred segments */

void BuildSegmentTree(SEGMENTHEADER *SegmentHeaders, ULONG numSegments) {
	ULONG j;
	ULONG k;
	ULONG l;
	SEGMENTHEADER *pSEG;

	for (j=0;j<numSegments;j++) {
		
		SegmentHeaders[j].numRefSymbolDict = 0;
		SegmentHeaders[j].numRefCustTables = 0;
		SegmentHeaders[j].bRefIntermediate = 0;
		SegmentHeaders[j].bRefHalfToneDict = 0;

		if (SegmentHeaders[j].ReferredCount) {
			SegmentHeaders[j].apSDSegments = jb2calloc(SegmentHeaders[j].ReferredCount,sizeof(*SegmentHeaders[j].apSDSegments));
			SegmentHeaders[j].apCTSegments = jb2calloc(SegmentHeaders[j].ReferredCount,sizeof(*SegmentHeaders[j].apCTSegments));
			for (k=0;k<SegmentHeaders[j].ReferredCount;k++) {
				pSEG = 0;
				for (l=0;l<j;l++)
					if (SegmentHeaders[l].Number==SegmentHeaders[j].aReferredNumbers[k]) {
						pSEG = &SegmentHeaders[l];
						break;
					}
				if (pSEG==0) {
					LogErrorV("Segment Refers to Unknown Segment ",SegmentHeaders[j].Number);
				} else
					if ((SegmentHeaders[j].PageNum != pSEG->PageNum)&&(pSEG->PageNum!=0))
						LogWarningV("Segment Refers to Segment on Different Page ",SegmentHeaders[j].Number);
					
					switch (pSEG->Type) {
						case 0: /* Symbol Dictionary */
							SegmentHeaders[j].apSDSegments[SegmentHeaders[j].numRefSymbolDict] = pSEG;
							SegmentHeaders[j].numRefSymbolDict++;
							break;
						case 16: /* HalfTone Dictionary */
							if (SegmentHeaders[j].bRefHalfToneDict) {
								LogErrorV("Segment Refers to More than One Halftone Dictionary ",SegmentHeaders[j].Number);
							} else {
								SegmentHeaders[j].bRefHalfToneDict = 1;
								SegmentHeaders[j].pHTDSegment = pSEG;
							}
							break;
						case 53: /* Tables */
							SegmentHeaders[j].apCTSegments[SegmentHeaders[j].numRefCustTables] = pSEG;
							SegmentHeaders[j].numRefCustTables++;
							break;
						case 4: /* Intermediate Symbol Region */
						case 20: /* Intermediate HalfTone Region */
						case 36: /* Intermediate Generic Regions */
						case 40: /* Intermediate refinement region */
							if (SegmentHeaders[j].bRefIntermediate) {
								LogErrorV("Segment Refers to More than One Intermediate Segment",SegmentHeaders[j].Number);
							} else {
								SegmentHeaders[j].bRefIntermediate = 1;
								SegmentHeaders[j].pINTSegment = pSEG;
							}
							break;
						default:		
							LogWarningV("Ignored: Segment Refers to Unused Segment",SegmentHeaders[j].Number);
					}
			}
		}
	}
}

/* Determine the length of the bitstream by looking for the End marker */

void SkipUnknownSegmentLength(DECODER *pDecoder) {

	unsigned short Buffer;
	unsigned short BufferStop;
	
	DecodeBits(pDecoder,136);
	Buffer = (BYTE) DecodeByte(pDecoder);
	if (getBIT(Buffer,0)) /* MMR */
		BufferStop = 0x0000;
	else {
		BufferStop = 0xFFAC;
		/* read past AT bits */
		if (getBITS(Buffer,1,2)==0)
			DecodeBits(pDecoder,64);
		else
			DecodeBits(pDecoder,16);
	}

	Buffer = 0x1111;

	while (Buffer!=BufferStop)
		Buffer = (unsigned short)((Buffer << 8) + (BYTE) DecodeByte(pDecoder));
}

/* Decode a Custom Huffman Table */

void DecodeTableSegment(SEGMENTHEADER *Seg) {

	ULONG flags;
	ULONG HTPS;
	ULONG HTRS;
	SLONG CURRANGELOW;
	ULONG NTEMP;
	BYTE *pDecStart;
	ULONG RL;

	DECODER *pDec = &Seg->pCoderData->Dec;

	Seg->CustTable.pDef = jb2malloc(sizeof(*Seg->CustTable.pDef));

	flags = DecodeByte(pDec);

	Seg->CustTable.pDef->HTOOB = getBIT(flags,0);
	HTPS = getBITS(flags,1,3) + 1;
	HTRS = getBITS(flags,4,6) + 1;
	Seg->CustTable.pDef->HTLOW = DecodeQuad(pDec);
	Seg->CustTable.pDef->HTHIGH = DecodeQuad(pDec);

	CURRANGELOW = Seg->CustTable.pDef->HTLOW;
	NTEMP = 0;

	pDecStart = DecoderPtr(pDec);

	/* Note: 2 Passes of the Table... Marginally Inefficient, but Simple */

	do {
		DecodeBits(pDec,HTPS);
		RL = DecodeBits(pDec,HTRS);
		CURRANGELOW += (1 << RL);
		NTEMP = NTEMP + 1;
	} while (CURRANGELOW < Seg->CustTable.pDef->HTHIGH);

	if (Seg->CustTable.pDef->HTOOB)
		NTEMP += 3;
	else
		NTEMP += 2;

	Seg->CustTable.PREFLEN = jb2malloc(NTEMP *sizeof(*Seg->CustTable.PREFLEN));
	Seg->CustTable.RANGELEN = jb2malloc(NTEMP *sizeof(*Seg->CustTable.RANGELEN));
	Seg->CustTable.RANGELOW = jb2malloc(NTEMP *sizeof(*Seg->CustTable.RANGELOW));
	Seg->CustTable.PREFCODES = jb2malloc(NTEMP * sizeof(*Seg->CustTable.PREFCODES));

	DecodeFlush(pDec);
	DecSkipBytes(pDec,pDecStart - DecoderPtr(pDec));

	CURRANGELOW = Seg->CustTable.pDef->HTLOW;
	NTEMP = 0;

	do {
		Seg->CustTable.PREFLEN[NTEMP] = DecodeBits(pDec,HTPS);
		Seg->CustTable.RANGELEN[NTEMP] = DecodeBits(pDec,HTRS);
		
		Seg->CustTable.RANGELOW[NTEMP] = CURRANGELOW;
		CURRANGELOW = CURRANGELOW + (1 << Seg->CustTable.RANGELEN[NTEMP]);
		NTEMP = NTEMP + 1;
	} while (CURRANGELOW < Seg->CustTable.pDef->HTHIGH);

	Seg->CustTable.pDef->numLines = NTEMP;

	Seg->CustTable.pDef->LOWPREFLEN = DecodeBits(pDec,HTPS);
	Seg->CustTable.PREFLEN[NTEMP] = Seg->CustTable.pDef->LOWPREFLEN;
	Seg->CustTable.RANGELEN[NTEMP] = 32;
	Seg->CustTable.RANGELOW[NTEMP] = Seg->CustTable.pDef->HTLOW - 1;
	NTEMP = NTEMP + 1;

	Seg->CustTable.pDef->HIGHPREFLEN = DecodeBits(pDec,HTPS);
	Seg->CustTable.PREFLEN[NTEMP] = Seg->CustTable.pDef->HIGHPREFLEN;
	Seg->CustTable.RANGELEN[NTEMP] = 32;
	Seg->CustTable.RANGELOW[NTEMP] = Seg->CustTable.pDef->HTHIGH;
	NTEMP =	NTEMP + 1;

	if (Seg->CustTable.pDef->HTOOB) {
		Seg->CustTable.pDef->OOBPREFLEN = DecodeBits(pDec,HTPS);
		Seg->CustTable.PREFLEN[NTEMP] = Seg->CustTable.pDef->OOBPREFLEN;
		Seg->CustTable.RANGELEN[NTEMP] = 0;
		NTEMP = NTEMP + 1;
	}

	HuffGetPrefCodes(Seg->CustTable.PREFLEN,Seg->CustTable.PREFCODES,NTEMP);
	BuildHuffTree(&Seg->CustTable.Tree,Seg->CustTable.PREFCODES,Seg->CustTable.PREFLEN,NTEMP);

}

/* Decode the configuration for a region segment */

void DecodeRegionSegment(DECODER *pDecoder, REGIONSEGMENT *RegSeg, BOOLN bUnknownHeight) {
	
	ULONG flags;

	RegSeg->Width = DecodeQuad(pDecoder);
	if (bUnknownHeight) {
		DecodeQuad(pDecoder); /* bitbucket? */
	} else
		RegSeg->Height = DecodeQuad(pDecoder);
	RegSeg->OffsetX = DecodeQuad(pDecoder);
	RegSeg->OffsetY = DecodeQuad(pDecoder);
	flags = DecodeByte(pDecoder);
	RegSeg->COMBPOP = getBITS(flags,0,2);
}

/* Using the proper masking rules, add a decoded region to the page */

void RegionToPage(SEGMENTHEADER *Seg) {

	ULONG op;
	BMP newpage;

	if (Seg->pPageSeg->bOverrideOP)
		op = Seg->RegSegment.COMBPOP;
	else
		op = Seg->pPageSeg->COMBPOP;

	if ((Seg->Type==42)||(Seg->Type==43))
		if (Seg->bRefIntermediate==0)
			op = COMBOPREPLACE;
	
	if (Seg->pPageSeg->bStriped) {
		if ((Seg->RegSegment.OffsetY + Seg->RegSegment.Height)>(Seg->pPageImage->Height)) {
			CreateBMP(&newpage,Seg->pPageImage->Width,(Seg->RegSegment.OffsetY + Seg->RegSegment.Height),Seg->pPageSeg->DefPixel);
			CombineBMP(&newpage,Seg->pPageImage,0,0,COMBOPREPLACE);
			DestroyBMP(Seg->pPageImage);
			*Seg->pPageImage = newpage;
		}
	}

	CombineBMP(Seg->pPageImage,&Seg->Region,Seg->RegSegment.OffsetX,Seg->RegSegment.OffsetY,op);

}

/* Decode the Start of Page Segment */

void DecodePageSegment(SEGMENTHEADER *Seg) {
	ULONG flags;
	DECODER Dec;
	
	InitDecoder(&Dec,Seg->pData,Seg->DataLength);
	
	Seg->pPageSeg->Width = DecodeQuad(&Dec);
	Seg->pPageSeg->Height = DecodeQuad(&Dec);
	Seg->pPageSeg->ResolutionX = DecodeQuad(&Dec);
	Seg->pPageSeg->ResolutionY = DecodeQuad(&Dec);

	flags = DecodeByte(&Dec);
	Seg->pPageSeg->bLossless = getBIT(flags,0);
	Seg->pPageSeg->bRefine = getBIT(flags,1);
	Seg->pPageSeg->DefPixel = getBIT(flags,2);
	Seg->pPageSeg->COMBPOP = getBITS(flags,3,4);
	Seg->pPageSeg->bAuxBuffers = getBIT(flags,5);
	Seg->pPageSeg->bOverrideOP = getBIT(flags,6);

	flags = DecodeBits(&Dec,16);
	Seg->pPageSeg->MaxStripeSize = getBITS(flags,0,14);
	Seg->pPageSeg->bStriped = getBIT(flags,15);

	if ((Seg->pPageSeg->Height==0xFFFFFFFF)&&(Seg->pPageSeg->bStriped==0))
		LogWarning("Unknown Height Requires Stripe");
}

/* Read in the Header structure for a segment */

void ReadSegmentHeader(DECODER *pDecoder, SEGMENTHEADER **SegmentHeaders,ULONG *numSegments,ULONG *maxSegments) {

	ULONG  Buff;
	ULONG j;
	ULONG PageFieldSize;
	SEGMENTHEADER *pSeg;

	if (*numSegments==*maxSegments)
		ReSizeArray((void **) SegmentHeaders,sizeof(**SegmentHeaders),maxSegments);

	pSeg = &(*SegmentHeaders)[*numSegments];

	pSeg->Number= DecodeQuad(pDecoder);

	Buff = DecodeByte(pDecoder);

	pSeg->Type = getBITS(Buff,0,5);
	if (getBIT(Buff,6))
		PageFieldSize = 32;
	else
		PageFieldSize = 8;

	pSeg->bDeferredNonRetain = getBIT(Buff,7);

	Buff = DecodeByte(pDecoder);

	if (getBITS(Buff,5,7)<=4) {

		pSeg->bRetainThis = getBIT(Buff,0);
		pSeg->ReferredCount = getBITS(Buff,5,7);
		if (pSeg->ReferredCount > 0) {
			pSeg->abReferredRetainFlags = jb2malloc(pSeg->ReferredCount * sizeof(*pSeg->abReferredRetainFlags));
			for (j=0;j<pSeg->ReferredCount;j++)
				pSeg->abReferredRetainFlags[j] = getBIT(Buff,j+1);
		}

	} else {
		
		pSeg->ReferredCount = getBITS(Buff,0,4) << 24;
		pSeg->ReferredCount += DecodeBits(pDecoder,24);
		if (pSeg->ReferredCount > 0) {
			pSeg->abReferredRetainFlags = jb2malloc(pSeg->ReferredCount * sizeof(*pSeg->abReferredRetainFlags));
			for (j=0;j<pSeg->ReferredCount;j++) {
				if (MOD8(j)==0) {
					Buff = DecodeByte(pDecoder);
					if (j==0)
						pSeg->bRetainThis = getBIT(Buff,0);
				}
				pSeg->abReferredRetainFlags[j] = getBIT(Buff,MOD8(j)+1);
			}
		}
	}

	if (pSeg->ReferredCount > 0) {
		pSeg->aReferredNumbers = jb2malloc(pSeg->ReferredCount * sizeof(long));
		for (j=0;j<pSeg->ReferredCount;j++)
			if (pSeg->Number<= 256)
				pSeg->aReferredNumbers[j] = DecodeByte(pDecoder);
			else if (pSeg->ReferredCount <= 655536)
				pSeg->aReferredNumbers[j] = DecodeBits(pDecoder,16);
			else
				pSeg->aReferredNumbers[j] = DecodeQuad(pDecoder);
	}

	pSeg->PageNum = DecodeBits(pDecoder,PageFieldSize);

	pSeg->DataLength = DecodeQuad(pDecoder);

	if (pSeg->DataLength==0xFFFFFFFF)
		pSeg->bUnknownLength = 1;
	else
		pSeg->bUnknownLength = 0;

	(*numSegments) ++;

}

/* Free any memory allocated for a segment */

void FreeSegmentHeader(SEGMENTHEADER *pSeg) {

	ULONG j;

	if (pSeg->ReferredCount > 0) {
		jb2free(pSeg->abReferredRetainFlags);
		jb2free(pSeg->aReferredNumbers);
		jb2free(pSeg->apSDSegments);
		jb2free(pSeg->apCTSegments);
	}

	switch (pSeg->Type) {
		case 0: /* Symbol Dictionary */
			if (pSeg->SymbolDict.Size > 0) {
				for (j=0;j<pSeg->SymbolDict.Size;j++)
					DestroyBMP(&pSeg->SymbolDict.aSymbols[j]);
				jb2free(pSeg->SymbolDict.aSymbols);
			}
			if (pSeg->bStoreContextData) {
				DestroyMQBuffer (&pSeg->MQGeneric);
				DestroyMQBuffer (&pSeg->MQRefine);
			}
			break;
		case 16: /* HalfTone Dictionary */
			if (pSeg->HTTable.Size > 0) {
				for (j=0;j<pSeg->HTTable.Size;j++)
					DestroyBMP(&pSeg->HTTable.aPatterns[j]);
				jb2free(pSeg->HTTable.aPatterns);
			}
			break;
		case 53: /* Tables */
			if (pSeg->CustTable.Tree != 0)
				jb2free(pSeg->CustTable.pDef);
				FreeHuffTable(&pSeg->CustTable);
			break;
		
		case 4: /* Intermediate Symbol Region */
		case 20: /* Intermediate HalfTone Region */
		case 36: /* Intermediate Generic Regions */
		case 40: /* Intermediate refinement region */
			if (pSeg->Region.Bits != NULL) {
				DestroyBMP (&pSeg->Region);
				LogWarning ("Intermediate Region was not used\n");
			}
			break;
	}
	
}
