/*

	huffman.c - Huffman routines for the Power JBIG-2 Decoder / Encoder


	Author:
		Dave Tompkins (dave.tompkins@ieee.org)
		University of British Columbia

	Source:
		http://spmg.ece.ubc.ca/jbig2
	  
	Revision History:

		1.0.0	1999-03-15	First Release


	Copyright:
		(c) 1998 University of British Columbia and Image Power, Inc.

		The program may be used for academic, demonstration and research 
		purposes only.  It may not be sold for profit or incorporated into
		any commercial programs and/or services without first obtaining a
		commercial useage license from Image Power, Inc.  This program is
		provided as is, without any express or implied warranty, including
		the warranty of fitness for a particular purpose or warranties of
		merchantability.

*/


#include "huffman.h"
#include "jb2common.h"

/* Define the standard huffman tables */

static HUFFTABLEDEF HuffmanStandardDefs[NUMHUFFSTANDARD] = {
{ 0,	0,		65808,	0,	3,	0,	3 },
{ 1,	0,		75,		0,	6,	6,	5 },
{ 1,	-256,	75,		8,	7,	6,	6 },
{ 0,	1,		76,		0,	5,	0,	5 },
{ 0,	-255,	76,		7,	6,	0,	6 },
{ 0,	-2048,	2048,	6,	6,	0,	12 },
{ 0,	-1024,	2048,	5,	5,	0,	13 },
{ 1,	-15,	1670,	9,	9,	2,	18 },
{ 1,	-31,	3339,	9,	9,	2,	19 },
{ 1,	-21,	4166,	8,	8,	2,	18 },
{ 0,	1,		141,	0,	7,	0,	12 },
{ 0,	1,		73,		0,	8,	0,	12 },
{ 0,	1,		140,	0,	7,	0,	12 },
{ 0,	-2,		3,		0,	0,	0,	5 },
{ 0,	-24,	25,		7,	7,	0,	11 } };

static ULONG HuffmanStandardPrefLen[] = {
1,	2,	3,
1,	2,	3,	4,	5,
8,	1,	2,	3,	4,	5,
1,	2,	3,	4,	5,
7,	1,	2,	3,	4,	5,
5,	4,	4,	4,	5,	5,	4,	2,	3,	3,	4,	4,
4,	3,	4,	5,	5,	4,	4,	5,	5,	4,	3,	3,	3,
8,	9,	8,	9,	7,	4,	2,	5,	6,	3,	6,	4,	4,	5,	5,	6,	7,	6,
8,	9,	8,	9,	7,	4,	3,	3,	5,	6,	3,	6,	4,	4,	5,	5,	6,	7,	6,
7,	8,	7,	5,	2,	5,	6,	7,	8,	2,	5,	6,	6,	6,	6,	6,	6,	7,
1,	2,	4,	4,	5,	5,	6,	7,	7,	7,	7,	7,
1,	2,	3,	5,	5,	6,	7,	7,	7,	7,	7,	8,
1,	3,	4,	5,	4,	3,	6,	6,	6,	6,	6,	7,
3,	3,	1,	3,	3,
7,	6,	5,	4,	3,	1,	3,	4,	5,	6,	7
};

static ULONG HuffmanStandardRangeLen[] = {
4,	8,	16,
0,	0,	0,	3,	6,
8,	0,	0,	0,	3,	6,
0,	0,	0,	3,	6,
8,	0,	0,	0,	3,	6,
10,	9,	8,	7,	6,	5,	5,	7,	7,	8,	9,	10,
9,	8,	7,	6,	5,	5,	5,	5,	6,	7,	8,	9,	10,
3,	1,	1,	0,	0,	0,	1,	0,	0,	4,	1,	4,	5,	6,	7,	7,	8,	10,
4,	2,	2,	1,	1,	1,	1,	1,	1,	1,	5,	2,	5,	6,	7,	8,	8,	9,	11,
4,	0,	0,	0,	2,	0,	0,	0,	0,	6,	5,	5,	6,	7,	8,	9,	10,	11,
0,	1,	0,	1,	1,	2,	2,	2,	3,	4,	5,	6,
0,	0,	1,	0,	1,	1,	0,	1,	2,	3,	4,	5,
0,	0,	0,	0,	1,	3,	1,	2,	3,	4,	5,	6,
0,	0,	0,	0,	0,
4,	2,	1,	0,	0,	0,	0,	0,	1,	2,	4
};

/* Given a huffman table & prefix lengths, determines the prefix codes */

void HuffGetPrefCodes(ULONG *aPrefLen, ULONG *aPrefixLenRefCodes, ULONG numCodes) {

	/* Note: assumes aPrefixLenRefCodes has already been initialized to size numCodes */

	ULONG LENMAX;
	ULONG CURLEN;
	ULONG CURCODE;
	ULONG CURTEMP;
	ULONG *LENCOUNT;
	ULONG *FIRSTCODE;
	ULONG ulLoop;

	LENMAX = 0;

	for (ulLoop=0;ulLoop<numCodes;ulLoop++) {
		aPrefixLenRefCodes[ulLoop] = 0;
		if (LENMAX<aPrefLen[ulLoop])
			LENMAX=aPrefLen[ulLoop];
	}

	LENCOUNT = jb2calloc((LENMAX+1),sizeof(*LENCOUNT));
	FIRSTCODE = jb2calloc((LENMAX+1),sizeof(*FIRSTCODE));

	for (ulLoop=0;ulLoop<numCodes;ulLoop++)
		if (aPrefLen[ulLoop]<=LENMAX)
			LENCOUNT[aPrefLen[ulLoop]]++;
		else
			LogError("Unexpected: HuffGetPrefCodes");

	CURLEN = 1;
	FIRSTCODE[0] = 0;
	LENCOUNT[0] = 0;

	while (CURLEN <= LENMAX) {
		FIRSTCODE[CURLEN] = (FIRSTCODE[CURLEN-1]+LENCOUNT[CURLEN-1]) * 2;
		CURCODE = FIRSTCODE[CURLEN];
		CURTEMP = 0;
		while (CURTEMP<numCodes) {
			if (aPrefLen[CURTEMP]==CURLEN) {
				aPrefixLenRefCodes[CURTEMP] = CURCODE;
				CURCODE ++;
			}
			CURTEMP++;
		}
		CURLEN++;
	}

	jb2free(LENCOUNT);
	jb2free(FIRSTCODE);
}

/* Frees a Huffman Table */

void FreeHuffTable(HUFFTABLE *pTable) {
	jb2free(pTable->PREFLEN);
	jb2free(pTable->RANGELEN);
	jb2free(pTable->RANGELOW);
	jb2free(pTable->PREFCODES);
	FreeHuffTree(pTable->Tree);
}

/* Builds & allocates memory for a huffman table from the basic definition */

void BuildHUFFTABLE(HUFFTABLE *pTable,ULONG *aucPref,ULONG *aucRange) {
	
	ULONG ulLoop,ulNumRows;
	SLONG CURRANGELOW;

	ulNumRows = pTable->pDef->numLines + 2;
	if (pTable->pDef->HTOOB)
		ulNumRows ++;

	pTable->PREFLEN = jb2malloc(ulNumRows * sizeof(*(pTable->PREFLEN)));
	pTable->RANGELEN = jb2malloc(ulNumRows * sizeof(*(pTable->RANGELEN)));
	pTable->RANGELOW = jb2malloc(ulNumRows * sizeof(*(pTable->RANGELOW)));
	pTable->PREFCODES = jb2malloc(ulNumRows * sizeof(*(pTable->PREFCODES)));

	CURRANGELOW = pTable->pDef->HTLOW;

	for (ulLoop=0;ulLoop<pTable->pDef->numLines;ulLoop++) {
		pTable->PREFLEN[ulLoop] = aucPref[ulLoop];
		pTable->RANGELEN[ulLoop] = aucRange[ulLoop];
		pTable->RANGELOW[ulLoop] = CURRANGELOW;
		CURRANGELOW += 1 << (pTable->RANGELEN[ulLoop]);
	}

	pTable->PREFLEN[pTable->pDef->numLines] = pTable->pDef->LOWPREFLEN;
	pTable->RANGELEN[pTable->pDef->numLines] = 32;
	pTable->RANGELOW[pTable->pDef->numLines] = pTable->pDef->HTLOW - 1;

	pTable->PREFLEN[pTable->pDef->numLines+1] = pTable->pDef->HIGHPREFLEN;
	pTable->RANGELEN[pTable->pDef->numLines+1] = 32;
	pTable->RANGELOW[pTable->pDef->numLines+1] = CURRANGELOW;

	if (pTable->pDef->HTOOB) {
		pTable->PREFLEN[pTable->pDef->numLines+2] = pTable->pDef->OOBPREFLEN;
		pTable->RANGELEN[pTable->pDef->numLines+2] = 0;
	}

	HuffGetPrefCodes(pTable->PREFLEN,pTable->PREFCODES,ulNumRows);

	BuildHuffTree(&pTable->Tree,pTable->PREFCODES,pTable->PREFLEN,ulNumRows);
}

/* Frees the Standard Huffman tables */

void FreeHuffStandard(HUFFTABLE *HuffStd) {
	ULONG ucLoop;
	for (ucLoop=0;ucLoop<NUMHUFFSTANDARD;ucLoop++) {
		FreeHuffTable(&HuffStd[ucLoop]);
	}
}

/* Builds the array of standard huffman tables */

void InitializeHuffStandard(HUFFTABLE *HuffStd) {
	ULONG ucLoop;
	ULONG *aucPref;
	ULONG *aucRange;

	aucPref = HuffmanStandardPrefLen;
	aucRange = HuffmanStandardRangeLen;

	for (ucLoop=0;ucLoop<NUMHUFFSTANDARD;ucLoop++) {
		HuffStd[ucLoop].pDef = &HuffmanStandardDefs[ucLoop];
		BuildHUFFTABLE(&HuffStd[ucLoop],aucPref,aucRange);
		aucPref += HuffStd[ucLoop].pDef->numLines;
		aucRange += HuffStd[ucLoop].pDef->numLines;
	}
}

/* Given a huffman table and a number to encode, determine prefix & Range bits */

void EncodeHuffNum(HUFFTABLE *pTable, SLONG slNum, ULONG *ulPrefixCode, ULONG *ulPrefixLen, ULONG *ulRangeCode, ULONG *ulRangeLen) {

	SLONG ulLoop;

	*ulPrefixLen = *ulRangeLen = 0;
	*ulPrefixCode = *ulRangeCode = 0;

 	if (slNum <= pTable->RANGELOW[pTable->pDef->numLines]) {
		if (pTable->PREFLEN[pTable->pDef->numLines]==0)
			LogError("EncodeHuffNum: Below HTLOW");
		else {
			*ulPrefixLen = pTable->PREFLEN[pTable->pDef->numLines];
			*ulPrefixCode = pTable->PREFCODES[pTable->pDef->numLines];
			*ulRangeLen = pTable->RANGELEN[pTable->pDef->numLines];
			*ulRangeCode = pTable->RANGELOW[pTable->pDef->numLines] - slNum;
		}
	} else if (slNum >= pTable->RANGELOW[pTable->pDef->numLines+1]) {
		if (pTable->PREFLEN[pTable->pDef->numLines+1]==0)
			LogError("EncodeHuffNum: Above HTHIGH");
		else {
			*ulPrefixLen = pTable->PREFLEN[pTable->pDef->numLines+1];
			*ulPrefixCode = pTable->PREFCODES[pTable->pDef->numLines+1];
			*ulRangeLen = pTable->RANGELEN[pTable->pDef->numLines+1];
			*ulRangeCode = slNum - pTable->RANGELOW[pTable->pDef->numLines+1];
		}
	} else {
		for (ulLoop=pTable->pDef->numLines-1;ulLoop>=0;ulLoop--)
			if (slNum >= pTable->RANGELOW[ulLoop]) {
				*ulPrefixLen = pTable->PREFLEN[ulLoop];
				*ulPrefixCode = pTable->PREFCODES[ulLoop];
				if (*ulPrefixLen==0)
					LogError("EncodeHuffNum: No Prefix Bits");
				*ulRangeLen = pTable->RANGELEN[ulLoop];
				*ulRangeCode = slNum - pTable->RANGELOW[ulLoop];
				break;
			}
	}

	if (*ulRangeLen+*ulPrefixLen==0)
		LogError("EncodeHuffNum: Length = 0");

}

/* returns the OOB prefix code for a huffman table */

void EncodeHuffOOB(HUFFTABLE *pTable, ULONG *ulPrefixCode, ULONG *ulPrefixLen) {

	if (pTable->pDef->HTOOB) {
		*ulPrefixLen = pTable->PREFLEN[pTable->pDef->numLines+2];
		*ulPrefixCode = pTable->PREFCODES[pTable->pDef->numLines+2];
	} else
		LogError("EncodeHuffOOB: Table Does Not Support OOB");
}

/* Creates a new branch in a huffman decode tree */

void NewHuffNode(HUFFNODE **newnode) {
	*newnode = jb2calloc(sizeof(**newnode),1);
}

/* Frees a huffman decode tree */

void FreeHuffTree(HUFFNODE *node) {
	if (node->p1)
		FreeHuffTree(node->p1);
	if (node->p0)
		FreeHuffTree(node->p0);
	jb2free(node);
}

/* Builds a Huffman Decoder Tree from the base structure */

void BuildHuffTree(HUFFNODE **mainnode, ULONG *PrefCodes, ULONG *PrefLens, ULONG numCodes) {

	ULONG j;
	ULONG k;
	HUFFNODE *pCurNode;

	NewHuffNode(mainnode);

	for (j=0;j<numCodes;j++)
		if (PrefLens[j]>0) {
			pCurNode = *mainnode;
			for (k=0;k<PrefLens[j];k++) {
				if (getBIT(PrefCodes[j],PrefLens[j]-k-1)) {
					if (pCurNode->p1==0)
						NewHuffNode(&pCurNode->p1);
					pCurNode = pCurNode->p1;
				} else {
					if (pCurNode->p0==0)
						NewHuffNode(&pCurNode->p0);
					pCurNode = pCurNode->p0;
				}
			}
			if ((pCurNode->p0!=0)||(pCurNode->p1!=0))
				LogError("BuildHuffTree: Invalid Prefix Codes");
			pCurNode->bEND = 1;
			pCurNode->NodeID = j;
		}
}
