/*

	jb2common.c - Common routines for the Power JBIG-2 Decoder / Encoder


	Author:
		Dave Tompkins (dave.tompkins@ieee.org)
		University of British Columbia

	Source:
		http://spmg.ece.ubc.ca/jbig2
	  
	Revision History:

		1.0.0	1999-03-15	First Release


	Copyright:
		(c) 1998 University of British Columbia and Image Power, Inc.

		The program may be used for academic, demonstration and research 
		purposes only.  It may not be sold for profit or incorporated into
		any commercial programs and/or services without first obtaining a
		commercial useage license from Image Power, Inc.  This program is
		provided as is, without any express or implied warranty, including
		the warranty of fitness for a particular purpose or warranties of
		merchantability.

*/


#include "jb2common.h"

/* An array used for determining strip sizes */

BYTE INVLOG2[4] = {1,2,4,8};
BYTE LOG2[9] = {0,0,1,0,2,0,0,0,3};

/* Dynamically resizes (doubles) an array if the ArraySize is reached */

void ReSizeArray(void **pArray, size_t ElementSize, ULONG *ArraySize) {
	*pArray = jb2realloc(*pArray,ElementSize * (*ArraySize) * 2);
	*ArraySize *= 2;
}

/* Initializes MQ Contexts */

void ResetMQBuffer(MQBUFFER *pMQB,ULONG BuffSize) {
	ULONG j;
	for (j=0;j<BuffSize;j++)
		pMQB->Index[j] = pMQB->MPS[j] = 0;
}

/* Allocates memory for an MQ Context Buffer */

void CreateMQBuffer(MQBUFFER *pMQB, ULONG BuffSize) {
	pMQB->Index = jb2calloc(BuffSize,sizeof(BYTE));
	pMQB->MPS = jb2calloc(BuffSize,sizeof(BYTE));
}

/* Frees memory for an MQ Context Buffer */

void DestroyMQBuffer (MQBUFFER *pMQB) {
	jb2free(pMQB->Index);
	jb2free(pMQB->MPS);
}

/* Allocates memory for MQ Integer Context Buffers */

void CreateMQInt(MQINTSTRUCT *MQInts, ULONG SYMCODELEN, void *pMQCoder) {
	ULONG j;
	ULONG IAIDArraySize;

	for (j=0;j<NUMMQINT;j++) {
		MQInts[j].pMQEnc = pMQCoder;
		MQInts[j].pMQDec = pMQCoder;
	}

	for (j=0;j<NUMMQINT-1;j++) {
		CreateMQBuffer(&MQInts[j].Buffer,512);
		MQInts[j].bIAID = 0;
	}

	MQInts[MQIAID].SBSYMCODELEN = SYMCODELEN;
	IAIDArraySize = 1 << SYMCODELEN;
	MQInts[MQIAID].bIAID = 1;
	CreateMQBuffer(&MQInts[MQIAID].Buffer,IAIDArraySize);
}

/* Frees memory for MQ Integer Context Buffers */

void DestroyMQInt(MQINTSTRUCT *MQInts) {
	ULONG j;
	for (j=0;j<NUMMQINT;j++)
		DestroyMQBuffer(&MQInts[j].Buffer);
}

/* A (very inefficient) routine used to check TPR for refinement regions */

BOOLN Check3x3(BMP *bmp,SLONG x, SLONG y) {

	BOOLN ret = 1;
	BOOLN prevval;
	SLONG j,k;
	
	prevval = GetBitBMPsafe(bmp,x-1,y-1);

	for (k=-1;k<2;k++)
		for (j=-1;j<2;j++)
			if (ret)
				if (prevval != GetBitBMPsafe(bmp,x+j,y+k))
					ret=0;
	return(ret);
}

/* Determine if a FileHeader of a .jb2 file is valid */

BOOLN ValidFileHeader(BYTE FileHeader[]) {

	BOOLN valid = 1;

	if (FileHeader[0]!=0x97) valid = 0;
	if (FileHeader[1]!=0x4A) valid = 0;
	if (FileHeader[2]!=0x42) valid = 0;
	if (FileHeader[3]!=0x32) valid = 0;
	if (FileHeader[4]!=0x0D) valid = 0;
	if (FileHeader[5]!=0x0A) valid = 0;
	if (FileHeader[6]!=0x1A) valid = 0;
	if (FileHeader[7]!=0x0A) valid = 0;

	return(valid);

}

/* if JB2MEM is turned on, some "container" routines for allocating memory.
Can be quite usefull if an implementation has special memory concerns, */

#ifdef JB2MEM

	SLONG memcount = 0;

	void *jb2malloc( size_t size ) {
		void *tmp;
		memcount++;
		tmp = malloc(size);
		if (tmp==NULL)
			LogError("Unable to Allocate Memory! Unexpected Results");
		return(tmp);
	}

	void *jb2calloc( size_t num, size_t size ) {
		void *tmp;
		memcount++;
		tmp = calloc(num,size);
		if (tmp==NULL)
			LogError("Unable to Allocate Memory! Unexpected Results");
		return(tmp);
	}

	void *jb2realloc( void *memblock, size_t size ) {
		void *tmp;
		tmp = realloc(memblock,size);
		if (tmp==NULL)
			LogError("Unable to Re-Allocate Memory! Unexpected Results");
		return(tmp);
	}


	void jb2free( void *memblock ) {
		memcount--;
		free(memblock);
	}

	SLONG NumAllocated() {
		return(memcount);
	}

#endif


