/*

	mqcontext.c - MQ Context routines used by the Power JBIG-2 Decoder / Encoder

	Author:
		Dave Tompkins (dave.tompkins@ieee.org)
		University of British Columbia

	Source:
		http://spmg.ece.ubc.ca/jbig2
	  
	Revision History:

		1.0.0	1999-03-15	First Release


	Copyright:
		(c) 1998 University of British Columbia and Image Power, Inc.

		The program may be used for academic, demonstration and research 
		purposes only.  It may not be sold for profit or incorporated into
		any commercial programs and/or services without first obtaining a
		commercial useage license from Image Power, Inc.  This program is
		provided as is, without any express or implied warranty, including
		the warranty of fitness for a particular purpose or warranties of
		merchantability.

*/


#include "jb2common.h"
#include "mqcontext.h"

/* Read in bytes from the image */

static void NextContextByte(CONTEXTLINE *pL) {

	if (pL->bEOL)
		pL->Count=65536;
	else {
		pL->nextxbyte++;
		if (pL->nextxbyte>0) {
			pL->Buff += (*pL->pnext);
			if (pL->nextxbyte - pL->img->MinWidthBytes==0) {
				if (MOD8(pL->img->Width)!=0)
					pL->Buff &= (BUFFMASK << (8 - MOD8(pL->img->Width)));
				pL->bEOL = 1;
			}
		}
		pL->pnext++;
		pL->Count=8;
	}
}


/* Setup for AT pixels */

static void InitContextATLine(MQCONTEXT *pC, BMP *img, SLONG deltax, ULONG posn, SLONG y) {
	CONTEXTATLINE *pATLin;
	pATLin = &pC->ATlin[pC->numATLines];
	pATLin->img = img;
	pATLin->DeltaX = deltax;
	pATLin->NextY = y;
	pATLin->Shift = posn + 16;
	pATLin->NextX = deltax;
	pC->numATLines++;
}

/* Setup for non-AT pixels */

static void InitContextLine(MQCONTEXT *pC, BMP *img, SLONG deltax, ULONG len, ULONG posn, SLONG y) {

	CONTEXTLINE *pLine;

	pLine = &pC->lin[pC->numLines];
	pLine->img = img;
	pLine->deltax = deltax;
	pLine->Mask = BUFFMASK << (BUFFLEN - len);
	pLine->Shift = posn + 16;
	pLine->nexty = y;
	pC->numLines++;
}

/* Setup a context structure, based on the template settings */

void InitContext(MQCONTEXT *pC, BMP *img,ULONG TEMPLATE, SLONG *ATX, SLONG *ATY) {

	pC->numLines=0;
	pC->numATLines = 0;

	switch (TEMPLATE) {
	case 0:
		InitContextATLine(pC,img,ATX[1],5,ATY[1]);
		InitContextLine(pC,img,-2,5,6,-1);
		InitContextATLine(pC,img,ATX[0],11,ATY[0]);
		InitContextATLine(pC,img,ATX[3],0,ATY[3]);
		InitContextLine(pC,img,-1,3,1,-2);
		InitContextATLine(pC,img,ATX[2],4,ATY[2]);
		pC->CurLineMask = BUFFMASK >> (BUFFLEN - 4);
		break;
	
	case 1:
		InitContextLine(pC,img,-1,4,3,-2);

		InitContextLine(pC,img,-2,5,7,-1);
		InitContextATLine(pC,img,ATX[0],12,ATY[0]);
		pC->CurLineMask = BUFFMASK >> (BUFFLEN - 3);
		break;

	case 2:

		InitContextLine(pC,img,-1,3,6,-2);
		InitContextLine(pC,img,-2,4,9,-1);
		InitContextATLine(pC,img,ATX[0],13,ATY[0]);
		pC->CurLineMask = BUFFMASK >> (BUFFLEN - 2);
		break;
	case 3:
		InitContextLine(pC,img,-3,5,6,-1);
		InitContextATLine(pC,img,ATX[0],11,ATY[0]);
		pC->CurLineMask = BUFFMASK >> (BUFFLEN - 4);
		break;

	}
}

/* Setup a reference context structure, based on the template settings */

void InitRefContext(MQCONTEXT *pC, BMP *source,BMP *dest, SLONG ox, SLONG oy, ULONG TEMPLATE, SLONG *ATX, SLONG *ATY) {

	pC->numLines=0;
	pC->numATLines = 0;
	
	switch (TEMPLATE) {
	case 0:
		InitContextLine(pC,dest,0,2,13,-1);
		InitContextATLine(pC,dest,ATX[0],12,ATY[0]);
		InitContextLine(pC,source,0-ox,2,4,-1-oy);
		InitContextATLine(pC,source,ATX[1]-ox,3,ATY[1]-oy);
		InitContextLine(pC,source,-1-ox,3,6,0-oy);
		InitContextLine(pC,source,-1-ox,3,9,1-oy);

		pC->CurLineMask = BUFFMASK >> (BUFFLEN - 1);
		break;
	
	case 1:
		InitContextLine(pC,dest,-1,3,12,-1);
		InitContextLine(pC,source,0-ox,1,6,-1-oy);
		InitContextLine(pC,source,-1-ox,3,7,0-oy);
		InitContextLine(pC,source,0,2,10,1-oy);

		pC->CurLineMask = BUFFMASK >> (BUFFLEN - 1);
		break;

	}
}


/* For each row of the image, re-set the Context data */

void NextContextRow(MQCONTEXT *pC) {
	ULONG j;
	SLONG k;
	CONTEXTLINE *pLine;
	for (j=0;j<pC->numLines;j++) {
		pLine = &pC->lin[j];

		pLine->Buff = 0;
		pLine->bEOL=0;
		pLine->Count = 0;
		
		if ((pLine->nexty<0)||(pLine->nexty>=(SLONG) pLine->img->Height)||(pLine->deltax > (SLONG) pLine->img->Width))
			pLine->bEOL = 1;
		else {
			pLine->nextxbyte = pLine->deltax / 8;
			if ((pLine->deltax < 0)&&(MOD8(pLine->deltax)!=0))
				pLine->nextxbyte--;

			pLine->pnext = pLine->img->Bits;
			pLine->pnext += (pLine->nexty * pLine->img->AlignWidthBytes + pLine->nextxbyte);

			NextContextByte(pLine);
			pLine->Buff <<= 8;
			NextContextByte(pLine);
			pLine->Buff <<= 8;
			NextContextByte(pLine);
			pLine->Buff <<= 8;
			NextContextByte(pLine);

			k = MOD8(pLine->deltax);
			if (k<0)
				k += 8;
			pLine->Buff <<= k;
			pLine->Count -= k;
		}
		pLine->nexty ++;
	}
	for (j=0;j<pC->numATLines;j++) {
		pC->ATlin[j].NextX = pC->ATlin[j].DeltaX;
		pC->ATlin[j].NextY++;
	}
	pC->CurLineBuff = 0;
}

/* Return the current context */

ULONG GetContext(MQCONTEXT *pC, BOOLN prevbit) {

	register CONTEXTLINE *pLine;
	register ULONG ret=0;
	ULONG j;

	for (j=0;j<pC->numLines;j++) {
		pLine = &pC->lin[j];
		ret += (pLine->Buff & pLine->Mask) >> pLine->Shift;
		pLine->Buff <<= 1;
		if ((pLine->Count -= 1)==0)
			NextContextByte(pLine);
	}

	for (j=0;j<pC->numATLines;j++) {
		ret += ((ULONG)GetBitBMPsafe(pC->ATlin[j].img,pC->ATlin[j].NextX,pC->ATlin[j].NextY - 1) << 31) >> pC->ATlin[j].Shift;
		pC->ATlin[j].NextX++;
	}

	ret += (pC->CurLineBuff = ((pC->CurLineBuff << 1) + prevbit) & pC->CurLineMask);
	
	return(ret);
}
